/*
** Copyright (C) 2003 The Free Software Foundation.
**  
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 2 of the License, or
** (at your option) any later version.
**  
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**  
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software
** Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
**  
*/


#include "libraries.h"
#include "utilities.h"
#include "ssf.h"
#include "book.h"
#include "bible.h"
#include "usfmtools.h"


Book::Book(int dummy) {
}


Book::~Book() {
}


void Book::clear() {
  path.clear();
  cleaned_lines.clear();
  previously_presented_chapter = 0;
}


void Book::load(const string& Path, Stylesheet& sheet, bool do_not_split_lines, const string& backupdirectory) {
  /*
     Clears any previous data, and loads the new book.
     It looks at the book in file "Path", and provides information on it.
  */
  // Clear up current data (if there is any)
  cleaned_lines.clear();
  // The stylesheet is needed to find out about the USFMs.
  stylesheet = sheet;
  path = Path;
  // Load, clean and split the lines.
  // The way to do it depends on whether we are loading a text file or a compressed file.
  string s;
  if (Glib::str_has_suffix(Path, ".gz")) {
    FILE *stream;
    s = "gunzip -c '";
    s.append (Path);
    s.append ("'");
    stream = popen(s.c_str(), "r");
    char buf[10240];
    while (fgets (buf, sizeof(buf), stream)) {
      s = buf;
      split_clean_lines (s, cleaned_lines, do_not_split_lines);
    }
    pclose(stream);
  } else {
    // Plain file, just load it.
    ReadText rt(path, true);
    for (unsigned int i = 0; i < rt.lines.size(); i++) {
      s = rt.lines[i];
      split_clean_lines (s, cleaned_lines, do_not_split_lines);
    }
  }
  // Go through the book to find out the number of chapters in it.
  number_of_chapters = 0;
  for (unsigned int i = 0; i < cleaned_lines.size(); i++) {
    s = cleaned_lines[i];
    s = usfm_extract_marker(s);
    if (stylesheet.usfm_is_chapter(s)) number_of_chapters++;
  }
  // Correct for somme one-chapter book do not have the c marker at all.
  if (number_of_chapters == 0) number_of_chapters++;

  // Location for saving the backups of this book.
  backup_directory = backupdirectory;
}


unsigned int Book::number_of_verses (unsigned int chapternumber) {
  /*
    This function returns the number of verses that is found
    in chapter "chapternumber".
  */
  unsigned int begin, end;
  get_chapter_boundaries (chapternumber, begin, end);
  unsigned int numberofverses = 0;
  for (unsigned int i = begin; i <= end; i++) {
    string line = cleaned_lines[i];
    string usfm = usfm_extract_marker (line);
    if (stylesheet.usfm_is_verse(usfm))
      numberofverses++;
  }
  return numberofverses;
}


void Book::split_clean_lines (const string& text, vector<string>& lines, bool do_not_split) {
  /*
Variable "text" brings in one line of text. This line is supposed to start with
an usfm. However, other usfms can be found within the line. There are
usfms that have also a closer marker. These can be left in the line. But the other
usfms cause the line of text to be split in multiple lines, which
are stored in variabele "lines".
However, if variabele "do_not_split" is true, the splitting is not done.
   */
  unsigned int position = 0;
  vector<unsigned int> cutpositions;
  string usfm;
  int index;
  position = usfm_search_marker_from (text, position);
  while (position != string::npos) {
    if (position > 0) {
      // A marker has been found, which is not at the beginning of the line.
      usfm = text.substr(position + 1, usfm_search_marker_from_length (text, position) - 1);
      index = stylesheet.UsfmIndex(usfm);
      // Ensure that an unknown marker does not cut the line.
      if (index >= 0) {
	// Ensure that a marker with an endmarker does not cut the line.
	if (stylesheet.GetEndmarker(index).empty()) {
	  cutpositions.push_back(position);
	}
      }
    }
    position = usfm_search_marker_from (text, ++position);
  }
  // At this point there is a vector that may either be empty or have cutting points.
  // If it is empty, no cuttings need be made, but if it has points, the line 
  // needs to be divided in pieces.
  // Variabele "do_not_split", if set, empties that vector with cutting positions.
  if (do_not_split) cutpositions.clear();
  string s;
  unsigned int previous_point = 0;
  for (unsigned int i = 0; i < cutpositions.size(); i++) {
    s = text.substr (previous_point, cutpositions[i] - previous_point);
    lines.push_back(trim(s));
    previous_point = cutpositions[i];
  }
  s = text.substr (previous_point, text.length());
  lines.push_back(trim(s));
}


void Book::get_chapter (unsigned int chapternumber, vector<string>& lines) {
  /*
    This function presents chapter "chapternumber" in "lines".
  */
  lines.clear();
  unsigned int begin, end;
  get_chapter_boundaries (chapternumber, begin, end);
  for (unsigned int i = begin; i <= end; i++) {
    lines.push_back(cleaned_lines[i]);
  }
  previously_presented_chapter = chapternumber; // Used later to accept the chapter again.
}


void Book::set_chapter (vector<string>& lines) {
  /*
    This function accepts the chapter in "lines", 
    and incorporates it again in the whole chapter,
    taking it as being the chapter number that was before presented.
  */
  // Get the two boundaries in the vector of strings, where the
  // chapter begins and where it ends.
  unsigned int begin, end;
  get_chapter_boundaries (previously_presented_chapter, begin, end);
  // Using an temporal vector of lines, start filling it with
  // 1. the set of lines before the current chapter starts.
  // 2. the current edited chapter.
  // 3. the lines following the current chapter.
  vector<string> temporal_lines;
  for (unsigned int i = 0; i < cleaned_lines.size(); i++) {
    if ((i < begin) || (i > end)) {
      temporal_lines.push_back(cleaned_lines[i]);
    } else {
      i = end;
      for (unsigned int i2 = 0; i2 < lines.size(); i2++) {
	temporal_lines.push_back(lines[i2]);
      }
    }
  }
  // Copy the temporal lines back to the book.
  cleaned_lines.clear();
  for (unsigned int i = 0; i < temporal_lines.size(); i++) {
    cleaned_lines.push_back(temporal_lines[i]);
  }
}


bool Book::save() {
  // Saves the book, and returns true if all went well.
  bool all_went_well = false;
  // Only save when we have got a filename.
  if (!path.empty()) {
    // Make backup of current file, if a directory is given where to put them.
    if (!backup_directory.empty()) {
      // Get the date and time in the right format.
      FILE *stream;
      stream = popen("date +%G-%m-%d-%H:%M-", "r");
      char buf[1024];
      fgets (buf, sizeof(buf), stream);
      string date = buf;
      date = trim(date);
      pclose(stream);
      // Get the filename to copy, and the path where to put it.
      string filename = Glib::path_get_basename(path);
      filename = Glib::build_filename (backup_directory, date + filename);
      // Carry out the copy command.
      string command = "cp '" + path + "' '" + filename + "'";
      system (command.c_str());
      // Compress the backup file. Force it to overwrite an existing file.
      command = "gzip -f '" + filename + "'";
      system (command.c_str());
    }
    try {
      WriteText wt(path);
      for (unsigned int i = 0; i < cleaned_lines.size(); i++) {
	wt.out << cleaned_lines[i] << endl;
      }
      all_went_well = true;
    }
    catch (int) {
      cerr << "Error writing to " << path << endl;
    }
  }
  return all_went_well;
}


void Book::get_chapter_boundaries (unsigned int chapter_number, unsigned int& begin, unsigned int& end) {
  /*
    This function finds the boundaries of the chapter number requested.
   */
  // Initialize the begin and end.
  begin = 0;
  end = 0;
  // Now go through all the text and start looking for the beginning and the end of the requested chapter.
  unsigned int CurrentChapter = 0;
  bool look_for_first_marker = true;
  for (unsigned int i = 0; i < cleaned_lines.size(); i++) {
    string s = cleaned_lines[i];
    s = usfm_extract_marker(s);
    if (CurrentChapter == 0) {
      // Allow for one-chapter books that have no \c 1 marker.
      if (stylesheet.usfm_is_verse(s)) CurrentChapter++;
    }
    if (stylesheet.usfm_is_chapter(s)) CurrentChapter++;
    if (CurrentChapter == chapter_number) {
      end = i;
      if (look_for_first_marker) {
	begin = i;
	look_for_first_marker = false;
      } 
    }
    // Speed up trick: break out when we have what we want.
    if (CurrentChapter > chapter_number) break;
  }
  // The first chapter also will include any headings. Care for that here.
  if (chapter_number == 1) 
    begin = 0;
}


vector<string> Book::get_backups () {
  /*
    Gets all the backups that exist for this book.
  */
  // This holds the values.
  vector<string> backups;
  // Get all the names.
  ReadDirectory rd (backup_directory, "", "");
  for (unsigned int i = 0; i < rd.files.size(); i++) {
    backups.push_back(rd.files[i]);
  }
  return backups;
}


string Book::get_backup_directory() {
  return backup_directory;
}


void Book::dump_with_references(vector<string>& lines) {
  /*
    This function dumps the whole book.
    Every line is preceded by a reference, e.g. MAT 1:1.
   */
  // Get the ID of the book, e.g. MAT.
  string id;
  for (unsigned int i = 0; i < cleaned_lines.size(); i++) {
    string line = cleaned_lines[i];
    string usfm = usfm_extract_marker(line);
    if (stylesheet.usfm_is_id(usfm)) {
      id = line.substr(0, 3);
      break;
    }
  }
  // Starting values for chapter and verse.
  string currentchapter = "1";
  string currentverse = "1";
  // Now go through all the lines, find out chapter and verse numbers,
  // and output the lines with the references prefixed.
  for (unsigned int i = 0; i < cleaned_lines.size(); i++) {
    string line = cleaned_lines[i];
    string usfm = usfm_extract_marker(line);
    if (stylesheet.usfm_is_chapter(usfm)) {
      currentchapter = number_in_string(line);
      currentverse = "1";
    } else if (stylesheet.usfm_is_verse(usfm)) {
      currentverse = number_in_string(line);
    }
    lines.push_back(id + " " + currentchapter + ":" + currentverse + " " + cleaned_lines[i]);
  }
}
