/*
** Copyright (C) 2003 The Free Software Foundation.
**  
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 2 of the License, or
** (at your option) any later version.
**  
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**  
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software
** Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
**  
*/


#include "mainwindow.h"
#include "libraries.h"
#include <gtkmm.h>
#include "dialogopen.h"
#include "dialogconvert2web.h"
#include "dialogshowscript.h"
#include "constants.h"
#include "config.h"
#include "utilities.h"
#include "ssf.h"
#include "help.h"
#include "usfmtools.h"
#include <myconfig.h>
#include "dialogsynchronization.h"
#include "synchronize.h"
#include "dialog_yes_no.h"
#include "dialogfind.h"
#include "log.h"
#include "dialogcompanion.h"


/*
|
|
|
|
|
Construction and destruction
|
|
|
|
|
 */


MainWindow::MainWindow() :
  menu_edit_undo("_Undo", Gtk::Menu_Helpers::AccelKey::AccelKey("<ctrl>z"), SigC::slot(*this, &MainWindow::menu_undo)),
  menu_edit_redo("_Redo", Gtk::Menu_Helpers::AccelKey::AccelKey("<ctrl><shift>z"), SigC::slot(*this, &MainWindow::menu_redo)),
  menuview_backups("View _backups", true),
  track(0),
  configuration(*this),
  scripture(0),
  book(0),
  backup(0),
  undo(0),
  windowscompanionlocal(0),
  windowscompanionremote(0)
{
  set_title("Bible Editor");
  set_icon_from_file(Glib::build_filename(DATAPATH, "icon.xpm"));

  // Size and position of the main window.
  resize(configuration.window_width, configuration.window_height);
  move (configuration.window_x_position, configuration.window_y_position);

  add(vbox);
  // We pack the menubar into the invisible box, which has been packed into the window.
  // Note that the pack_start default arguments are Gtk::EXPAND | Gtk::FILL, 0
  // The default is not what we want, therefore we enter 0. This causes the menubar not to expand at all,
  // neither to fill. It always stays at the top of the window.
  vbox.pack_start(menubar, 0, 0);
  {
    using namespace Gtk::Menu_Helpers;
    // Deal with the File menu.
    menubar.items().push_back(MenuElem("_File", menufile));
    menufile.items().push_back(StockMenuElem(Gtk::Stock::OPEN, SigC::slot(*this, &MainWindow::open))); 
    menufile.items().push_back(StockMenuElem(Gtk::Stock::CLOSE, SigC::slot(*this, &MainWindow::close)));
    menufile.items().push_back(MenuElem("Open _references", SigC::slot(*this, &MainWindow::on_open_references)));
    menufile.items().push_back(MenuElem("Save references _as", SigC::slot(*this, &MainWindow::on_save_references)));
    menufile.items().push_back(StockMenuElem(Gtk::Stock::QUIT, SigC::slot(*this, &MainWindow::hide)));
    // Deal with the Edit menu.
    menubar.items().push_back(MenuElem("_Edit", menuedit));
    menuedit.signal_show().connect( SigC::slot(*this, &MainWindow::menu_edit));
    menuedit.items().push_back(StockMenuElem(Gtk::Stock::CUT, SigC::slot(*this, &MainWindow::on_cut))); 
    menuedit.items().push_back(StockMenuElem(Gtk::Stock::COPY, SigC::slot(*this, &MainWindow::on_copy)));
    menuedit.items().push_back(StockMenuElem(Gtk::Stock::PASTE, SigC::slot(*this, &MainWindow::on_paste)));
    menuedit.items().push_back(menu_edit_undo);
    menuedit.items().push_back(menu_edit_redo);
    menuedit.items().push_back(StockMenuElem(Gtk::Stock::FIND, SigC::slot(*this, &MainWindow::menu_find)));    
    // Deal with the View menu.
    menubar.items().push_back(MenuElem("_View", menuview));
    menuview_backups.signal_toggled().connect( SigC::slot(*this, &MainWindow::on_view_backups));
    menuview.items().push_back(menuview_backups);
    // Deal with the Goto menu.
    menubar.items().push_back(MenuElem("_Goto", menugoto));
    menugoto.items().push_back(MenuElem("Next verse", AccelKey::AccelKey("F7"), SigC::slot(*this, &MainWindow::goto_next_verse)));
    menugoto.items().push_back(MenuElem("Previous verse", AccelKey::AccelKey("<shift>F7"), SigC::slot(*this, &MainWindow::goto_previous_verse)));
    menugoto.items().push_back(MenuElem("Next chapter", AccelKey::AccelKey("F8"), SigC::slot(*this, &MainWindow::goto_next_chapter)));
    menugoto.items().push_back(MenuElem("Previous chapter", AccelKey::AccelKey("<shift>F8"), SigC::slot(*this, &MainWindow::goto_previous_chapter)));
    menugoto.items().push_back(MenuElem("Next book", AccelKey::AccelKey("F9"), SigC::slot(*this, &MainWindow::goto_next_book)));
    menugoto.items().push_back(MenuElem("Previous book", AccelKey::AccelKey("<shift>F9"), SigC::slot(*this, &MainWindow::goto_previous_book)));
    // Deal with the Tools menu.
    menubar.items().push_back(MenuElem("_Tools", menutools));
    menutools.items().push_back(MenuElem("_Convert project to web pages", SigC::slot(*this, &MainWindow::convert2web)));
    menutools.items().push_back(MenuElem("_Synchronize projects now", AccelKey::AccelKey("<ctrl>S"), SigC::slot(*this, &MainWindow::synchronize_now)));
    menutools.items().push_back(MenuElem("Synchronize _interactively", SigC::slot(*this, &MainWindow::synchronize_interactive)));
    // Deal with the Preferences menu.
    menubar.items().push_back(MenuElem("_Preferences", menupreferences));
    menupreferences.items().push_back(MenuElem("_Synchronization with Paratext", SigC::slot(*this, &MainWindow::synchronization)));
    menupreferences.items().push_back(MenuElem("_Bibledit Windows Companion", SigC::slot(*this, &MainWindow::companion_settings)));
    menupreferences.items().push_back(MenuElem("_Editor font", SigC::slot(*this, &MainWindow::editor_settings)));
    // Deal with the Diagnostics menu.
    menubar.items().push_back(MenuElem("_Diagnostics", menudiagnostics));
    menudiagnostics.items().push_back(MenuElem("Show the output of the last _general script run", SigC::slot(*this, &MainWindow::showgeneralscript)));
    menudiagnostics.items().push_back(MenuElem("Show the output of the last _synchronization script run", SigC::slot(*this, &MainWindow::showsynchronizationscript)));
    menudiagnostics.items().push_back(MenuElem("_View system log", SigC::slot(*this, &MainWindow::viewlog)));
    // Deal with the Help menu.
    menubar.items().push_back(MenuElem("_Help", menuhelp));
    menuhelp.items().push_back(StockMenuElem(Gtk::Stock::HELP, SigC::slot(*this, &MainWindow::showhelp)));
    menuhelp.items().push_back(MenuElem("_About", SigC::slot(*this, &MainWindow::showabout)));
  }

  // Deal with the toolbar.
  vbox.pack_start(hbox, 0, 0);
  backbutton.add_pixlabel(Glib::build_filename(DATAPATH, "back.xpm"), "");
  backbutton.set_size_request(20);
  backbutton.signal_clicked().connect( SigC::slot(*this, &MainWindow::goto_previous_reference));
  hbox.pack_start(backbutton, 0, 0);
  forwardbutton.add_pixlabel(Glib::build_filename(DATAPATH, "forward.xpm"), "");
  forwardbutton.set_size_request(20);
  forwardbutton.signal_clicked().connect( SigC::slot(*this, &MainWindow::goto_next_reference));
  hbox.pack_start(forwardbutton, 0, 0);
  hbox.pack_start(combo_book, 0, 0);
  combo_book.get_entry()->signal_changed().connect( SigC::slot(*this, &MainWindow::on_combo_book_changed));
  combo_book.get_entry()->set_editable(false); // Not editable, else the user might enter junk, and crash the program.
  combo_chapter.set_size_request(55);
  hbox.pack_start(combo_chapter, 0, 0);
  combo_chapter.get_entry()->signal_changed().connect( SigC::slot(*this, &MainWindow::on_combo_chapter_changed));
  combo_chapter.get_entry()->set_editable(false);
  combo_verse.set_size_request(55);
  hbox.pack_start(combo_verse, 0, 0);
  combo_verse.get_entry()->signal_changed().connect( SigC::slot(*this, &MainWindow::on_combo_verse_changed));
  combo_verse.get_entry()->set_editable(false);

  // Deal with the editing area.
  hpane.set_position(configuration.hpane_position);
  vbox.pack_start(hpane);
  vpane.set_position(configuration.vpane_position);
  hpane.add1(vpane);

  // Deal with the list area.
  // Add the scrolled window to the right of the moveable pane.
  hpane.add2(scrolledwindow_list);
  // Only show the scrollbars when they are necessary.
  scrolledwindow_list.set_policy(Gtk::POLICY_AUTOMATIC, Gtk::POLICY_AUTOMATIC);
 //Add the treeview inside a scrolledwindow
  scrolledwindow_list.add(treeview);
  //Create the Tree model:
  ref_tree_model = Gtk::ListStore::create(columns);
  treeview.set_model(ref_tree_model);
  //Add the TreeView's view columns:
  treeview.append_column("References", columns.col_name);
  // Message handler.
  treeview.get_selection()->signal_changed().connect(SigC::slot(*this, &MainWindow::on_list_selection_changed));

  // Deal with the editor area.
  vpane.add1(scrolledwindow_text);
  scrolledwindow_text.set_policy(Gtk::POLICY_AUTOMATIC, Gtk::POLICY_AUTOMATIC);
  scrolledwindow_text.add(textview);
  textview.set_wrap_mode(Gtk::WRAP_WORD);
  on_cursor_moved_initialized = false;
  textview.signal_move_cursor().connect( SigC::slot(*this, &MainWindow::on_cursor_moved));
  textview.signal_grab_focus().connect( SigC::slot(*this, &MainWindow::on_grab_focus));
  textview.get_buffer()->signal_insert().connect( SigC::slot(*this, &MainWindow::on_insert));
  textview.get_buffer()->signal_erase().connect( SigC::slot(*this, &MainWindow::on_erase));
  fonttag = textview.get_buffer()->create_tag("font"); // To set the font properties.
  fonttag->property_font() = configuration.editor_font_name;
  
  // Deal with the notes area.
  vpane.add2(scrolledwindow_notes);
  scrolledwindow_notes.set_policy(Gtk::POLICY_AUTOMATIC, Gtk::POLICY_AUTOMATIC);
  scrolledwindow_notes.add(textview2);
  // We cannot change the text in the notes area.
  textview2.set_editable(false);
  textview2.set_wrap_mode(Gtk::WRAP_WORD);

  // Show the widgets. They will not really be shown until this window is shown.
  show_all_children();

  // Project initialization.
  initialize_project();

  // Get the system for tracking the Bible references up and running.
  Glib::signal_timeout().connect(SigC::slot(*this, &MainWindow::on_tracker_timeout), 1000);
  set_sensitivity_of_navigator_buttons();

  // System for navigation.
  Glib::signal_timeout().connect(SigC::slot(*this, &MainWindow::on_navigation_timeout), 100);

  // Automatic saving of the file, periodiodically.
  Glib::signal_timeout().connect(SigC::slot(*this, &MainWindow::on_save_timeout), 60000);

  // Automatic synchronization
  Glib::signal_timeout().connect(SigC::slot(*this, &MainWindow::on_synchronize_timeout), 60000);

  // Communication with Windows Companion
  if (configuration.companion_local_enabled)
    windowscompanionlocal.Connect("localhost", 51515);
  if (configuration.companion_remote_enabled)
    windowscompanionremote.Connect(configuration.companion_remote_name, 51515);
  Glib::signal_timeout().connect(SigC::slot(*this, &MainWindow::on_windows_companion_timeout), 100);

  // Entry in the log.
  log_message("Bibledit started");

}


MainWindow::~MainWindow() {
  // Delete script files and log files.
  unlink(temporary_file(SCRIPT_OUTPUT_FILE).c_str());
  unlink(Glib::build_filename(Glib::get_home_dir(), "unison.log").c_str());
  unlink(synchronize_temporary_filename().c_str());
  // Save the size and position of the program window.
  // The get_position functions is broken, See the gtkmm documentation for reasons.
  // Anyway, for just now we'll leave the variables in configuration, and we hope
  // that it will be fixed later. The problem now is that the function will
  // return the value that was set with move(x,y). If the user moves the window
  // that is not reflected here, but it will keep returning the values set
  // with move(x,y).
  // If you want to change the default values when Bibledit starts, edit
  // .configuration.
  get_size (configuration.window_width, configuration.window_height);
  get_position(configuration.window_x_position, configuration.window_y_position);
  configuration.hpane_position = hpane.get_position();
  configuration.vpane_position = vpane.get_position();
  // Save text in editor
  editor_save();
  // Remove system log.
  unlink(temporary_file(LOGFILE).c_str());
}


/*
|
|
|
|
|
Initialization
|
|
|
|
|
 */


void MainWindow::initialize_project() {

  // Find out which project is loaded.
  ScriptureSettingsFile ssf(configuration.projectdirectory);
  int projectindex = ssf.index(configuration.projectshortcut);
  if (projectindex < 0) {
    enable_or_disable_menus(false);
    return;
  }

  // Enable the menus.
  enable_or_disable_menus(true);

  // Load stylesheet.
  stylesheet.readsheet(ssf.StyleSheet(projectindex));

  // Load the data about the scripture.
  scripture.load(configuration.projectdirectory, projectindex);

  // Variable needed otherwise the navigation rumbles too much.
  driving_book = true;

  // Put the names of the books in the combo.
  vector<string> strings;
  string gotobook;
  for (unsigned int i = 0; i < scripture.IDs.size(); i++) {
    strings.push_back(scripture.IDs[i] + " " + scripture.EnglishBookNames[i]);
    if (i == 0) 
      gotobook = scripture.IDs[i] + " " + scripture.EnglishBookNames[i];
  }
  combo_book.set_popdown_strings(strings);
  int bookindex = scripture.get_index_of_id(configuration.book_opened);
  if (bookindex >= 0) {
    // Set the book to the one that was opened last.
    gotobook = scripture.IDs[bookindex] + " " + scripture.EnglishBookNames[bookindex];
  } else {
    // If the book does not exist, reset chapter and verse.
    configuration.chapter_opened = "1";
    configuration.verse_opened = "1";
  }

  // Combo has been filled: clear the anti-rumble variable.
  driving_book = false;

  // Clear some variables so that the new book will be loaded.
  current_book.clear();
  current_chapter.clear();
  current_verse.clear();

  // Go to the new reference.
  if (!gotobook.empty())
    go_to_reference (gotobook, configuration.chapter_opened, configuration.verse_opened, true);

  // Clear the references tracker.
  track.clear();
}


void MainWindow::enable_or_disable_menus(bool enable) {
  menugoto.set_sensitive(enable);
  menutools.set_sensitive(enable);
  menuedit.set_sensitive(enable);
  menuview.set_sensitive(enable);
  combo_book.set_sensitive(enable);
  combo_chapter.set_sensitive(enable);
  combo_verse.set_sensitive(enable);
  if (!enable) {
    combo_book.get_entry()->set_text("");
    combo_chapter.get_entry()->set_text("");
    combo_verse.get_entry()->set_text("");
  }
}


/*
|
|
|
|
|
Menu callbacks
|
|
|
|
|
 */


void MainWindow::open() {
  OpenDialog opendialog(*this, configuration);
  if (opendialog.run() == Gtk::RESPONSE_OK) {
    configuration.projectdirectory = opendialog.projectdirectory;
    configuration.projectshortcut = opendialog.projectshortcut;
    editor_save();
    initialize_project();
  }
}


void MainWindow::close() {
  editor_save();
  book.clear();
  enable_or_disable_menus(false);
  textview.get_buffer()->set_text("");
  current_book.clear();
  current_chapter.clear();
  current_verse.clear();
  requested_book.clear();
  requested_chapter.clear();
  requested_verse.clear();
  goto_book.clear();
  goto_chapter.clear();
  goto_verse.clear();
  configuration.projectshortcut.clear();
  ref_tree_model->clear();
}


void MainWindow::convert2web() {
  ConvertToWebDialog converttoweb(*this, configuration);
  if (converttoweb.run() == Gtk::RESPONSE_OK) {
    configuration.converttowebdirectory = converttoweb.converttowebdirectory;
  }
}


void MainWindow::showgeneralscript() {
  ShowScriptDialog showscript(*this, temporary_file(SCRIPT_OUTPUT_FILE));
  showscript.run();
}


void MainWindow::showsynchronizationscript() {
  ShowScriptDialog showscript(*this, synchronize_temporary_filename().c_str());
  showscript.run();
}


void MainWindow::viewlog() {
  ShowScriptDialog showscript(*this, temporary_file(LOGFILE));
  showscript.run();
}


void MainWindow::showhelp() {
  htmlhelp(*this);
}


void MainWindow::showabout() {
  string message;
  message = PACKAGE_STRING;
  message.append ("\n\n");
  message.append (PACKAGE_BUGREPORT);
  message.append ("\n");
  Gtk::MessageDialog dialog(*this, message);
  dialog.run();
}


void MainWindow::synchronization() {
  SynchronizationDialog synchronizationdialog(*this, configuration);
  if (synchronizationdialog.run() == Gtk::RESPONSE_OK) {
    configuration.synchronize_paratext_projectdirectory = synchronizationdialog.directory;
    configuration.synchronize_paratext = synchronizationdialog.checkbutton.get_active();
    {
    WriteText wt(synchronization_paths_filename());
    wt.out << synchronizationdialog.textview.get_buffer()->get_text();
    }
    synchronize_handle_config_file(configuration.synchronize_paratext, configuration.projectdirectory, configuration.synchronize_paratext_projectdirectory);
  }
}


void MainWindow::synchronize_now() {
  synchronize_execute();
}


void MainWindow::synchronize_interactive() {
  synchronize(true);
}


void MainWindow::menu_undo() {
  undo.undo(textview.get_buffer());
}


void MainWindow::menu_redo() {
  undo.redo(textview.get_buffer());
}


void MainWindow::menu_edit() {
  // Set the sensitivity of the items under the edit menu.
  menu_edit_undo.get_child()->set_sensitive(undo.can_undo());
  menu_edit_redo.get_child()->set_sensitive(undo.can_redo());
}


void MainWindow:: menu_find() {
  // Before finding, save the current file.
  editor_save();
  // Start find dialog.
  FindDialog finddialog(*this, configuration.projectdirectory, configuration.projectshortcut);
  finddialog.entry.set_text(configuration.find_searchword);
  if (finddialog.run() == Gtk::RESPONSE_OK) {
    configuration.find_searchword = finddialog.entry.get_text();
    menuview_backups.set_active(false);
    ReadText rt (temporary_file(FINDRESULTS), false);
    fill_list_store(rt.lines);
  }
}


void MainWindow::companion_settings() {
  CompanionDialog companiondialog (*this, configuration);
  if (companiondialog.run() == Gtk::RESPONSE_OK) {
    // If there is a change in the local connection, deal with it.
    if (configuration.companion_local_enabled != companiondialog.checkbuttonlocal.get_active()) {
      configuration.companion_local_enabled = companiondialog.checkbuttonlocal.get_active();
      if (configuration.companion_local_enabled) {
	windowscompanionlocal.Connect("localhost", 51515);
      } else {
	windowscompanionlocal.Disconnect();
      }
    }
    // Deal with any changes in the remote connection.
    bool b = false;
    if (configuration.companion_remote_enabled != companiondialog.checkbuttonremote.get_active()) 
      b = true;
    if (configuration.companion_remote_name.compare(companiondialog.entry.get_text()))
      b = true;
    if (b) {
      configuration.companion_remote_enabled = companiondialog.checkbuttonremote.get_active();
      configuration.companion_remote_name = companiondialog.entry.get_text();
      if (configuration.companion_remote_enabled) {
	windowscompanionremote.Connect(configuration.companion_remote_name, 51515);
      } else {
	windowscompanionremote.Disconnect();
      }
    }
  }
}


void MainWindow::editor_settings() {
  Gtk::FontSelectionDialog dialog;
  dialog.set_font_name(configuration.editor_font_name);
  int result = dialog.run();
  if (result == Gtk::RESPONSE_OK) {
    configuration.editor_font_name = dialog.get_font_name();
    fonttag->property_font() = configuration.editor_font_name;
  }
}


/*
|
|
|
|
|
Navigation
|
|
|
|
|
 */


void MainWindow::on_combo_book_changed() {
  if (driving_book) return;
  string book = combo_book.get_entry()->get_text();
  if (book.empty()) return;
  go_to_reference(book, "", "", false);
}


void MainWindow::on_combo_chapter_changed() {
  if (driving_chapter) return;
  string chapter = combo_chapter.get_entry()->get_text();
  if (chapter.empty()) return;
  go_to_reference("", chapter, "", false);
}


void MainWindow::on_combo_verse_changed() {
  if (driving_verse) return;
  string verse = combo_verse.get_entry()->get_text();
  if (verse.empty()) return;
  go_to_reference("", "", verse, false);
   
}


void MainWindow::goto_next_verse() {
  set_next_or_previous_item(combo_verse, combo_verse.get_entry()->get_text(), true);
}


void MainWindow::goto_previous_verse() {
  set_next_or_previous_item(combo_verse, combo_verse.get_entry()->get_text(), false);
}


void MainWindow::goto_next_chapter() {
  set_next_or_previous_item(combo_chapter, combo_chapter.get_entry()->get_text(), true);
}


void MainWindow::goto_previous_chapter() {
  set_next_or_previous_item(combo_chapter, combo_chapter.get_entry()->get_text(), false);
}


void MainWindow::goto_next_book() {
  set_next_or_previous_item(combo_book, combo_book.get_entry()->get_text(), true);
}


void MainWindow::goto_previous_book() {
  set_next_or_previous_item(combo_book, combo_book.get_entry()->get_text(), false);
}


void MainWindow::go_to_reference(const string& book, const string& chapter, const string& verse, bool jump) {
  /*
    This function makes the editor to go to a new Bible reference.

    Going to a reference is not done immediately on request. It is instead done by a scheduler.
    This is done because at times the program generates a lot of requests at once.
    If the edit control is then loaded each time, and the cursor positioned,
    this confuses Gtk when the requests come too soon one after the other.
    So this function only stores the requested references, and a timer will carry
    this request out later. 

    If the previously displayed reference was edited and therefore changed,
    it will save the file.

    The "jump" variabele needs to be set when any bigger jump is made from one to another reference.
    It is related to the interal logic for jumping.
   */

  requested_book.push_back(book);
  requested_chapter.push_back(chapter);
  requested_verse.push_back(verse);
  requested_jump.push_back(jump);
}


bool MainWindow::on_navigation_timeout() {
  /*
    This scheduler carries out requested changes of the reference.
  */

  // Carry out any pending repositioning of the cursor.
  // If F7, F8 or F9 is continually held down and it repeats,
  // we've code here to deal with that so that the inteface responds
  // properly. In cases like this the repositioning of the cursor
  // is delayed until the key is released.
  if (keypress_timer > 1) {
    if (!goto_cursor_position.empty()) {
      editor_position_cursor_at_verse(goto_cursor_position);
      goto_cursor_position.clear();
    }
  }

  // Take the reference we go to from the fifo.
  bool goto_jump;
  if (!requested_book.empty()) {
    string s;
    s = requested_book[0];
    requested_book.erase(requested_book.begin());
    if (!s.empty()) goto_book = s;
    s = requested_chapter[0];
    requested_chapter.erase(requested_chapter.begin());
    if (!s.empty()) goto_chapter = s;
    s = requested_verse[0];
    requested_verse.erase(requested_verse.begin());
    if (!s.empty()) goto_verse = s;
    goto_jump = requested_jump[0];
    requested_jump.erase(requested_jump.begin());
  }  

  // Find out what needs to be changed: book, chapter and/or verse.
  bool new_book = (goto_book != current_book);
  bool new_chapter = (goto_chapter != current_chapter);
  bool new_verse = (goto_verse != current_verse);

  // See whether the current file should be saved.
  if (new_book || new_chapter) {
    if (textview.get_buffer()->get_modified()) {
      editor_save();
    }
  }

  // A new book?
  if (new_book) {
    // Does the requested book exist?
    int bookindex = scripture.get_index_of_id(goto_book);
    if (bookindex >= 0) {
      // Adjust the current indicator, and store the book for later.
      goto_book = scripture.IDs[bookindex] + " " + scripture.EnglishBookNames[bookindex];
      current_book = goto_book;
      configuration.book_opened = goto_book;
      // Set the text in the combo box.
      driving_book = true;
      combo_book.get_entry()->set_text(goto_book);
      driving_book = false;
      // Load book, do not split the lines.
      book.load(scripture.FilePathAndNames[bookindex], stylesheet, true, scripture.get_backup_directory(bookindex));
      // Put the chapters in the chapter combo box.
      vector<string> strings;
      for (unsigned int i = 1; i <= book.number_of_chapters; i++) {
	strings.push_back(convert_to_string(i));
      }
      driving_chapter = true;
      combo_chapter.set_popdown_strings(strings);
      driving_chapter = false;
      // As we've a new book, go to chapter one verse one, except when "jumping".
      if (!goto_jump) {
	goto_chapter = "1";
	goto_verse = "1";
      }
      new_chapter = true;
      // If we are viewing the backups, update the view.
      view_backups_if_applicable();
    }
  }

  // Verify new chapter within limits
  if (new_chapter) {
    unsigned int int_goto_chapter = convert_to_int(goto_chapter);
    if (int_goto_chapter < 1) new_chapter = false;
    if (int_goto_chapter > book.number_of_chapters) new_chapter = false;
  }

  // A new chapter?
  if (new_chapter) {
    // Set the text in the combo box.
    driving_chapter = true;
    combo_chapter.get_entry()->set_text(goto_chapter);
    driving_chapter = false;
    // Store the current chapter.
    current_chapter = goto_chapter;
    configuration.chapter_opened = current_chapter;
    // Put the number of verses in combo_verse
    vector<string> strings;
    unsigned int chapter = convert_to_int (goto_chapter);
    if (chapter > 0) {
      unsigned int number_of_verses = book.number_of_verses(chapter);
      for (unsigned int i = 1; i <= number_of_verses; i++) {
	strings.push_back(convert_to_string(i));
      }
      driving_verse = true;
      combo_verse.set_popdown_strings(strings);
      driving_verse = false;
    }
    // Move to beginning of the chapter, except when "jumping"
    if (!goto_jump)
      goto_verse = "1";
    new_verse = true;
    // Load chapter, if need be.
    editor_load_chapter();
  }

  // Verify new verse within limits
  if (new_verse) {
    int int_goto_verse = convert_to_int(goto_verse);
    int limit = book.number_of_verses(convert_to_int(goto_chapter));
    if (int_goto_verse < 1) new_verse = false;
    if (int_goto_verse > limit) new_verse = false;
  }

  // A new verse?
  if (new_verse) {
    // Set the text in the combo box.
    driving_verse = true;
    combo_verse.get_entry()->set_text(goto_verse);
    driving_verse = false;
    // Cursor move;
    // Update the current verse, and store it for later also.
    current_verse = goto_verse;
    configuration.verse_opened = current_verse;
  }

  // Increase the keypress timer, see below for explanation.
  keypress_timer++;

  if (new_book || new_chapter || new_verse) {
    // Position the cursor properly.
    // The positioning will be done next time this routine is called. 
    // This is because sometimes Gtk is slow in loading a new chapter.
    // So if the cursor positioning is done straight after loading,
    // it might not work.
    goto_cursor_position = goto_verse;
    // Send the reference to the windows companions.
    string value = "gotoreference " + goto_book.substr(0, 3) + " " + goto_chapter + ":" + goto_verse;
    windowscompanionlocal.Sendln(value);
    windowscompanionremote.Sendln(value);
  }

  // Also carry out the change of the reference after the user has clicked somewhere in the text.
  on_grab_focus_timer++;
  if (on_grab_focus_timer == 1) {
    on_cursor_moved(Gtk::MOVEMENT_LOGICAL_POSITIONS, 0, false);
  }

  return true;
}


void MainWindow::on_cursor_moved(Gtk::MovementStep step, int count, bool extended_selection) {
  /*
    When the cursor has moved, the navigation system needs to be updated
    so that it shows the right reference. If the user was, for example
    on MAT 1:10, and the cursor moves, the move might have brought him or her
    to another reference, though this is not necessarily so. Therefore, as we 
    don't know where the user is now after the cursor moved, we need to find
    it out. The book is known, the chapter is known, because both stay the same.
    The only thing we don't know is the verse. 
   */
  if (!on_cursor_moved_initialized) {
    on_cursor_moved_initialized = true;
    return;
  }
  go_to_reference ("", "", cursor_get_verse_number(), false);
}


void MainWindow::on_grab_focus() {
  // The purpose of this is to adjust the reference to the new position,
  // wherever the user clicks in the text.
  on_grab_focus_timer = 0;
}


string MainWindow::cursor_get_verse_number() {
  /*
    This gets the verse number the cursor is on.
    We will find the verse out by doing the following things:
    1. Get the line the cursor is on. If it has a verse marker, we then can get
       the verse number from there.
    2. If we did not find a verse number, get the previous line, and see what
       information is there.
    3. Keep doing step 2 until we either have the information we want, or we have
       arrived at the beginning of the file.
    4. If we are at the start of the file, assume verse one.
   */
  // Get the line the cursor is on.
  Gtk::TextIter iterator (textview.get_buffer()->get_iter_at_mark(textview.get_buffer()->get_insert()));
  int line_of_cursor = iterator.get_line();
  // If no verse information is going to be found, assume verse one.
  string verse = "1";
  // Go through all lines, from the cursor downwards, and check whether verse information is there.
  Gtk::TextIter endofline;
  for (int i = line_of_cursor; i >= 0; i--) {
    iterator = textview.get_buffer()->get_iter_at_line(i);
    endofline = (textview.get_buffer()->get_iter_at_line(i+1));
    endofline--;
    string s = textview.get_buffer()->get_text(iterator, endofline);
    string usfm = usfm_extract_marker (s);
    if (stylesheet.usfm_is_verse(usfm)) {
      verse = number_in_string (s);
      break;
    }
  }
  return verse;
}


void MainWindow::editor_position_cursor_at_verse(const string& verse) {
  /*
    This function moves the cursor of the editor to "verse".
  */

  // We will not move the cursor or scroll to it when the cursor is already on the right verse
  if (verse.compare (cursor_get_verse_number())) {
    // Grab focus here to get the scrolling done properly, and the user can type in it.
    textview.grab_focus();
    // Assemble the marker to position the cursor at, e.g. "\v 1" for verse one.
    string marker;
    marker.append("\\");
    marker.append(stylesheet.get_verse_usfm());
    marker.append(" ");
    marker.append(verse);
    marker.append(" ");
    // Search for that marker.
    Gtk::TextIter iterator (textview.get_buffer()->begin());
    Gtk::TextIter iteratorbegin;
    Gtk::TextIter iteratorend (textview.get_buffer()->end());
    if (iterator.forward_search(marker, Gtk::TEXT_SEARCH_TEXT_ONLY, iteratorbegin, iteratorend, iteratorend)) {
      // If the marker is there, move the cursor to it.
      on_cursor_moved_initialized = false;
      textview.get_buffer()->place_cursor(iteratorend);
      // Scroll also to it. It will scroll to the beginning of the text after the verse marker.
      // Alignment is needed to put the line being edited near the top of the window.
      textview.scroll_to_iter(iteratorend, 0, 0, 0.1); 
    }
  }
}


/*
|
|
|
|
|
Tracker
|
|
|
|
|
 */


void MainWindow::goto_next_reference() {
  set_next_or_previous_reference(true);
}


void MainWindow::goto_previous_reference() {
  set_next_or_previous_reference(false);
}


bool MainWindow::set_next_or_previous_item(Gtk::Combo& combo, const string& current_text, bool next) {
  // This function takes the next (or previous) value in the dropdown list of the combo box.
  // It returns true if a next (or previous) value could be set, else false.
  vector<string> strings;
  strings = combo.get_popdown_strings();
  unsigned int index;
  string combo_text;
  for (unsigned int i = 0; i < strings.size(); i++) {
    if (current_text == strings[i])
      index = i;
  }
  if (next) {
    if (index == (strings.size() - 1)) return false;
    index++;
  } else {
    if (index == 0) return false;
    index--;
  }
  combo_text = strings[index];
  // Go to the new reference.
  if (&combo == &combo_book) go_to_reference (combo_text, "", "", false);
  if (&combo == &combo_chapter) go_to_reference ("", combo_text, "", false);
  if (&combo == &combo_verse) go_to_reference ("", "", combo_text, false);
  // We've changed a reference: reset the keypress_timer.
  keypress_timer = 0;
  // Return true as we've done the job.
  return true;
}


void MainWindow::set_sensitivity_of_navigator_buttons() {
  // Enables or disables each navigator button, depending on whether references for it are available.
  backbutton.set_sensitive(track.previous_reference_available());
  forwardbutton.set_sensitive(track.next_reference_available());
}


bool MainWindow::on_tracker_timeout() {
  // Drive the Bible reference tracker.
  track.tick(combo_book.get_entry()->get_text(), combo_chapter.get_entry()->get_text(), combo_verse.get_entry()->get_text());
  set_sensitivity_of_navigator_buttons();
  return true;
}


void MainWindow::set_next_or_previous_reference(bool next) {
  string book;
  string chapter;
  string verse;
  track.get_next_reference (book, chapter, verse, next);
  go_to_reference (book, chapter, verse, false);
  set_sensitivity_of_navigator_buttons();
}


/*
|
|
|
|
|
Editing
|
|
|
|
|
 */


void MainWindow::editor_load_chapter() {
  /*
    Loads the chapter in the editor.
  */
  // Set the new text.
  vector<string> lines;
  book.get_chapter(convert_to_int(combo_chapter.get_entry()->get_text()), lines);
  string text;
  for (unsigned int i = 0; i < lines.size(); i++) {
    text.append(lines[i] + "\n");
  }
  textview.get_buffer()->set_text(text);
  // Apply the right font.
  textview.get_buffer()->apply_tag (fonttag, textview.get_buffer()->begin(), textview.get_buffer()->end());
  // Set the editor ummodified.
  textview.get_buffer()->set_modified(false);
  // Clear the undo / redo stack.
  undo.clear();
}


void MainWindow::editor_save() {
  /*
    This will save the book if it has been changed in the editor.
   */
  // See whether it was changed.
  if (textview.get_buffer()->get_modified()) {
    // Collect all lines of the editor
    vector<string> lines;
    int number_of_lines = textview.get_buffer()->get_line_count();
    Gtk::TextIter iterator;
    Gtk::TextIter endofline;
    for (int i = 0; i < number_of_lines; i++) {
      iterator = textview.get_buffer()->get_iter_at_line(i);
      endofline = (textview.get_buffer()->get_iter_at_line(i+1));
      endofline--;
      string s = textview.get_buffer()->get_text(iterator, endofline);
      s = trim(s);
      if (!s.empty()) lines.push_back(s);
    }
    // Rejoin the lines in the chapter and save the whole book.
    book.set_chapter(lines);
    book.save();
    // Set the buffer to unmodified.
    textview.get_buffer()->set_modified(false);
    // If we are viewing the backups, update the view.
    view_backups_if_applicable();
  }
}


bool MainWindow::on_save_timeout() {
  editor_save();
  return true;
}


/*
|
|
|
|
|
Clipboard
|
|
|
|
|
 */


void MainWindow::on_cut() {
  Glib::RefPtr<Gtk::Clipboard> refClipboard = Gtk::Clipboard::get();
  if (textview.has_focus())
    textview.get_buffer()->cut_clipboard(refClipboard);
}


void MainWindow::on_copy() {
  Glib::RefPtr<Gtk::Clipboard> refClipboard = Gtk::Clipboard::get();
  if (textview.has_focus())
    textview.get_buffer()->copy_clipboard(refClipboard);
  if (textview2.has_focus())
    textview2.get_buffer()->copy_clipboard(refClipboard);
}


void MainWindow::on_paste() {
  Glib::RefPtr<Gtk::Clipboard> refClipboard = Gtk::Clipboard::get();
  if (textview.has_focus())
    textview.get_buffer()->paste_clipboard(refClipboard);
}


/*
|
|
|
|
|
Backups
|
|
|
|
|
 */


void MainWindow::on_view_backups() {
  if (menuview_backups.get_active()) {
    view_backups_if_applicable();
  } else {
    // Free resources used for backup.
    backup.clear();
    textview2.get_buffer()->set_text("");
    // Add new data.
    treeview.remove_all_columns();
    treeview.append_column("References", columns.col_name);
    vector<string> references;
    fill_list_store(references);
  }
}


void MainWindow::view_backups_if_applicable() {
  if (menuview_backups.get_active()) {
    treeview.remove_all_columns();
    treeview.append_column("Backups", columns.col_name);
    vector<string> files = book.get_backups();
    backup.load(book.get_backup_directory(), files);
    vector<string> names;
    for (unsigned int i = 0; i < backup.size(); i++) {
      names.push_back(backup.get_friendly_name(i));
    }
    fill_list_store(names);
    on_list_selection_changed();
  }
}

/*
|
|
|
|
|
List store and reference handling
|
|
|
|
|
 */


void MainWindow::fill_list_store(vector<string>& lines) {
  /*
    Fill the list store with the data.
   */
  // Clear the store first.
  ref_tree_model->clear();
  // Now add new data.
  Gtk::TreeModel::Row row;
  for (unsigned int i = 0; i < lines.size(); i++) {
    row = *(ref_tree_model->append());
    row[columns.col_name] = lines[i];
  }
}


void MainWindow::on_list_selection_changed() {
  /*
    Handler if the selection in the list changed
  */
  // Get the string the user selected.
  Gtk::TreeModel::iterator iter = treeview.get_selection()->get_selected();
  if(iter) {
    Gtk::TreeRow row = *iter;
    string value;
    row.get_value(0,value);
    // Do something depending on the backup / reference view.
    if (menuview_backups.get_active()) {
      // We're at the backup view.
      // Read the text of the right chapter of the book.
      Book book(0);
      book.load(backup.get_full_name(value), stylesheet, true, "");
      vector<string> lines;
      book.get_chapter (convert_to_int(current_chapter), lines);
      // Put the text in the buffer.
      string text;
      for (unsigned int i = 0; i < lines.size(); i++) {
	text.append(lines[i] + "\n");
      }
      textview2.get_buffer()->set_text(text);
    } else {
      // We're on the references view.
      // Extract the book.
      string book;
      if (value.length() > 3) 
	book = value.substr(0, 3);
      value.erase(0, 3);
      value = trim(value);
      // Extract chapter.
      string chapter;
      chapter = number_in_string(value);
      value.erase(0, chapter.length());
      value = trim(value);
      // Extract verse.
      string verse;
      verse = number_in_string(value);
      // Jump to the reference.
      go_to_reference(book, chapter, verse, true);
    }
  }
}


void MainWindow::on_open_references() {
  Gtk::FileSelection dialog("Please choose a file");
  dialog.set_transient_for(*this);
  dialog.hide_fileop_buttons();	
  dialog.set_filename(configuration.references_file);
  int result = dialog.run();
  if (result == Gtk::RESPONSE_OK) { 
    configuration.references_file = dialog.get_filename();
    ReadText rt(configuration.references_file, true);
    fill_list_store(rt.lines);
  }
}


void MainWindow::on_save_references() {

}


/*
|
|
|
|
|
Synchronization
|
|
|
|
|
 */


void MainWindow::synchronize_execute() {
  // Do not run it if another instance is already on.
  if (program_is_running("unison bibledit")) return;
  // Check previous output. If problem found, inform the user.
  if (string_in_file(synchronize_temporary_filename(), "<-?->")) {
    YesNoDialog ynd (*this, "Your help is needed with the synchronization\nDo you want to help now?");
    int result = ynd.run();
    if (result == Gtk::RESPONSE_YES) {
      synchronize(true);
      return;
    }
  }
  // Run the program.
  synchronize(false);
}


bool MainWindow::on_synchronize_timeout() {
  synchronize_execute();
  return true;
}


/*
|
|
|
|
|
Undo / Redo
|
|
|
|
|
 */


void MainWindow::on_insert(const Gtk::TextBuffer::iterator& pos, const Glib::ustring& text, int bytes) {
  undo.on_insert(pos, text, bytes);
}


void MainWindow::on_erase(const Gtk::TextBuffer::iterator& range_begin, const Gtk::TextBuffer::iterator& range_end) {
  undo.on_erase(range_begin, range_end);
}


/*
|
|
|
|
|
Bibledit Windows Companion
|
|
|
|
|
 */


bool MainWindow::on_windows_companion_timeout() {
  if (configuration.companion_local_enabled)
    windowscompanionlocal.Tick();
  if (configuration.companion_remote_enabled)
    windowscompanionremote.Tick();
  return true;
}


