/*
** Copyright (C) 2003 The Free Software Foundation.
**  
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 2 of the License, or
** (at your option) any later version.
**  
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**  
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software
** Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
**  
*/


#include "libraries.h"
#include "ssf.h"
#include "utilities.h"



ScriptureSettingsFile::ScriptureSettingsFile (const string& MyParatextProjects) {
  ProjectsDirectory = MyParatextProjects + "/";
  GetAllShortcuts();
}

ScriptureSettingsFile::~ScriptureSettingsFile () {
}


#define SSF_SUFFIX ".ssf"

#define LDS_SUFFIX1 ".lds"
#define LDS_SUFFIX2 ".LDS"


void ScriptureSettingsFile::GetAllShortcuts () {
  // Gets all shortcuts such as AVV, NDB, GRK, etc.
  ReadDirectory rd (ProjectsDirectory, "", SSF_SUFFIX);
  shortcuts.clear();
  string s;
  for (unsigned int i = 0; i < rd.files.size(); i ++) {
    s = rd.files[i];
    s = s.substr(0, s.length() - strlen(SSF_SUFFIX));
    shortcuts.push_back(s);
  }
}


string ScriptureSettingsFile::GetSSFData (unsigned int index, SSFDataType SSFData) {
  // Read one setting from the XML .ssf file.
  string returnvalue;
  string s;
  string s2;
  unsigned int i;
  s = ProjectsDirectory + "/" + shortcuts [index] + SSF_SUFFIX;
  ReadText rt(s);
  switch (SSFData) {
  case ssfdtFullName :
    s = "FullName";
    break;
  case ssfdtDirectory :
    s = "Directory";
    break;
  case ssfdtNaming :
    s = "Naming";
    break;
  case ssfdtFontName : 
    s = "DefaultFont";
    break;
  case ssfdtFontSize :
    s = "DefaultFontSize";
    break;
  case ssfdtVersification :
    s = "Versification";
    break;
  case ssfdtLanguage :
    s = "Language";
    break;
  case ssfdtStyleSheet :
    s = "StyleSheet";
    break;
  }
  for (i = 0; i < rt.lines.size(); i++) {
    s2 = trim(rt.lines[i]);
    if (s2.length() > 2) {
      if (s2.substr(1, s.length()) == s) {
	s2 = s2.substr(s.length() + 2, 1000);
	s2 = s2.substr(0, s2.length() - s.length() - 3);
	returnvalue = s2;
	break;
      }
    }
  }
  return returnvalue;
}


string ScriptureSettingsFile::GetLDSData (const string& Language, LDSDataType LDSData) {
  // Reads one setting from the .lds file.

  vector<string> v;
  string s;
  string s2;
  unsigned int i;
  string returnvalue;

  // It has been observed that some files have a lower case suffix, and others an uppercase. 
  // So check for both suffixes.
  s = ProjectsDirectory + "/" + Language + LDS_SUFFIX1;
  {
    ReadText rt(s, true);
    for (unsigned int i = 0; i < rt.lines.size(); i++) 
      v.push_back(rt.lines[i]);
  }
  s = ProjectsDirectory + "/" + Language + LDS_SUFFIX2;
  {
    ReadText rt(s, true);
    for (unsigned int i = 0; i < rt.lines.size(); i++) 
      v.push_back(rt.lines[i]);
  }
  switch (LDSData) {
  case ldsdtFontName :
    s = "font=";
    break;
  case ldsdtFontSize :
    s = "size=";
    break;
  }
  for (i = 0; i < v.size(); i++) {
    s2 = trim(v[i]);
    if (s2.length() > 2) {
      if (s2.substr(0, s.length()) == s) {
	returnvalue = s2.substr(s.length(), 1000);
	break;
      }
    }
  }
  return returnvalue;
}


string ScriptureSettingsFile::FullName (unsigned int index) {
  return GetSSFData (index, ssfdtFullName);
}


string ScriptureSettingsFile::Directory (unsigned int index) {
  /*
    This is a bit tricky to implement. That is because Paratext usually runs
    under Windows, and therefore the Directory that is read from the .ssf 
    file has usually the form of e.g.  C:\My Paratext Projects\NDB
    This path is useless under Unix. The solution (and I hope it always works,
    but that mostly depends on how consistent Paratext is with naming the
    directories), is this:
    1. The last part of the directory is looked at only, in this example NDB
    2. The Paratext Directory passed to this program is prefixed to that.
    3. The result is returned.
   */
  string s = GetSSFData (index, ssfdtDirectory);
  // Change all backslashes to forward slashes for the extraction of the filename to succeed.
  unsigned int i = s.find("\\");
  while (i != string::npos) {
    s.replace(i, 1, "/");
    i = s.find("\\");
  }
  s = Glib::path_get_basename(s);
  return ProjectsDirectory + s;
}


void ScriptureSettingsFile::Naming (unsigned int index, string& PrePart, string& PostPart, string& BookNameForm) {

#define PREPARTS "PrePart"
#define POSTPARTS "PostPart"
#define BOOKNAMEFORMS "BookNameForm"

  string s;
  string s2;
  unsigned int i;
  s = GetSSFData (index, ssfdtNaming);

  i = s.find(PREPARTS);
  if (i != string::npos) {
    s2 = s.substr(i + strlen(PREPARTS) + 2, 1000);
    i = s2.find ("\"");
    PrePart = s2.substr(0, i);
  } else {
    PrePart.clear();
  }

  i = s.find(POSTPARTS);
  if (i != string::npos) {
    s2 = s.substr(i + strlen(POSTPARTS) + 2, 1000);
    i = s2.find ("\"");
    PostPart = s2.substr(0, i);
  } else  {
    PostPart.clear();
  }

  i = s.find(BOOKNAMEFORMS);
  if (i != string::npos) {
    s2 = s.substr(i + strlen(BOOKNAMEFORMS) + 2, 1000);
    i = s2.find ("\"");
    BookNameForm = s2.substr(0, i);
  } else {
    BookNameForm.clear();
  }

#undef PREPARTS
#undef POSTPARTS
#undef BOOKNAMEFORMS
}


string ScriptureSettingsFile::FontName (unsigned int index) {
  return GetSSFData (index, ssfdtFontName);
}


unsigned int ScriptureSettingsFile::FontSize (unsigned int index) {
  string s;
  unsigned int returnvalue;
  s = number_in_string(GetSSFData (index, ssfdtFontSize));
  if (s.empty()) {
    returnvalue = 12;
  } else {
    returnvalue = convert_to_int (s);
  }
  return returnvalue;
}


unsigned int ScriptureSettingsFile::Versification (unsigned int index) {
  return convert_to_int (GetSSFData (index, ssfdtVersification));
}


string ScriptureSettingsFile::StyleSheet (unsigned int index) {
  /*
    This is a bit tricky, see under routine Directory why. 
    The same conversion that is made there is made here too.
   */
  string s = GetSSFData (index, ssfdtStyleSheet);
  // Change all backslashes to forward slashes for the extraction of the filename to succeed.
  unsigned int i = s.find("\\");
  while (i != string::npos) {
    s.replace(i, 1, "/");
    i = s.find("\\");
  }
  s = Glib::path_get_basename(s);
  return ProjectsDirectory + s;
}


string ScriptureSettingsFile::LDSFontName (unsigned int index) {
  string language;
  language = GetSSFData (index, ssfdtLanguage);
  return GetLDSData (language, ldsdtFontName);
}


unsigned int ScriptureSettingsFile::LDSFontSize (unsigned int index) {
  string language;
  language = GetSSFData (index, ssfdtLanguage);
  string size;
  size = GetLDSData (language, ldsdtFontSize);
  if (size.empty()) {
    return 12;
  } else {
    return convert_to_int(size);
  }
}


int ScriptureSettingsFile::index(const string& shortcut) {
  // Takes scripture shortcut, and returns its index in all the scriptures.
  int returnvalue = -1;
  for (unsigned int i = 0; i < shortcuts.size(); i++) {
    if (shortcut == shortcuts[i]) {
      returnvalue = i;
      break;
    }
  }
  return returnvalue;
}
