/*
** Copyright (C) 2003 The Free Software Foundation.
**  
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 2 of the License, or
** (at your option) any later version.
**  
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**  
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software
** Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
**  
*/


#include "libraries.h"
#include "utilities.h"
#include "stylesheet.h"



Stylesheet::Stylesheet() {
}


Stylesheet::~Stylesheet () {
}


void Stylesheet::readsheet (const string& filename) {
/*
Reads the lines of the stylesheet, and accumulates them.
When empty space is read, the accumulated lines are processed,
and so forth, until the end of the stylesheet is reached.
Before reading it clean all variables.
This is because this object may read different stylesheets
at different times.
*/
  cout << "Reading stylesheet " << filename << endl;
  // First clear all variables of a possible previously loaded stylesheet.
  ChapterUsfm.clear();
  IDUsfm.clear();
  RunningHeaderUsfm.clear();
  style_marker.clear();
  style_texttype.clear();
  style_textproperty_chapter.clear();
  style_textproperty_paragraph.clear();
  style_textproperty_publishable.clear();
  style_textproperty_nonpublishable.clear();
  style_textproperty_nonvernacular.clear();
  style_textproperty_vernacular.clear();
  style_textproperty_book.clear();
  style_textproperty_level.clear();
  style_textproperty_poetic.clear();
  style_textproperty_note.clear();
  style_textproperty_crossreference.clear();
  style_textproperty_verse.clear();
  style_styletype.clear();
  style_bold.clear();
  style_italic.clear();
  style_superscript.clear();
  style_endmarker.clear();
  style_justification.clear();
  style_first_line_indent.clear();
  style_fontsize.clear();
  style_leftmargin.clear();
  style_rightmargin.clear();
  style_space_after.clear();
  style_space_before.clear();
  // IMPORTANT: When adding new variables to the object, have them cleared here too.

  // Start loading this sheet.
  ReadText rt(filename, true);
  vector<string> lines;
  string line;
  for (unsigned int i = 0; i < rt.lines.size(); i++) {
    line = rt.lines[i];
    if (line.find("#") == 0) line.clear(); // Skip comments
    if (line.empty()) {
      LoadStylesheet (lines);
      lines.clear();
    } else {
      lines.push_back(line);
    }
  }
  LoadStylesheet (lines);
}


void Stylesheet::LoadStylesheet (vector<string>& style) {
  // Reads and stores the stylesheet.

  if (style.empty()) return;	// No styles, so quit.
  // Go through the style vectors one by one, and see if something to fill it
  // is in the style sheet.
  // If so, fill the slot, if not, fill in a default value.
  // This way the system can never be upset by either missing or double values.
  string s;

  s = usfm_get_style (style, "\\Marker");
  if (s.empty()) return; 	// No marker found, so incomplete style, so quit.
  style_marker.push_back(s);

  s = usfm_get_style (style, "\\TextType");
  if (s == "chapternumber") style_texttype.push_back(ttChapterNumber);
  else if (s == "notetext") style_texttype.push_back(ttNoteText);
  else if (s == "other") style_texttype.push_back(ttOther);
  else if (s == "section") style_texttype.push_back(ttSection);
  else if (s == "title" ) style_texttype.push_back(ttTitle);
  else if (s == "versenumber") style_texttype.push_back(ttVerseNumber);
  else if (s == "versetext") style_texttype.push_back(ttVerseText);
  // If no texttype was given, make it verse text, so that at least the content
  // does not disappear into the nothing, but gets displayed.
  else {
    style_texttype.push_back(ttVerseText);
    cerr << "Marker " << style_marker[style_marker.size() - 1] << " has no TextType, so the default VerseText has been assigned" << endl;
  }

  s = usfm_get_style (style, "\\TextProperties");
  // If searching for e.g. "publishable", and the properties contain 
  // "nonpublishable", then this would give a false positive. To prevent this
  // spaces are added to both sides of the string, and instead of searching
  // for "publishable" we now search for " publishable ". This would yield
  // the proper results.
  s.insert(0, " ");
  s.append(" ");

  style_textproperty_chapter.push_back((s.find (" chapter ") != string::npos));
  style_textproperty_paragraph.push_back((s.find (" paragraph ") != string::npos));
  style_textproperty_publishable.push_back((s.find (" publishable ") != string::npos));
  style_textproperty_nonpublishable.push_back((s.find (" nonpublishable ") != string::npos));
  style_textproperty_vernacular.push_back((s.find (" vernacular ") != string::npos));
  style_textproperty_nonvernacular.push_back((s.find (" nonvernacular ") != string::npos));
  style_textproperty_book.push_back((s.find (" book ") != string::npos));
  string nis = number_in_string(s);
  if (nis.length()) {
    unsigned int cc = convert_to_int(nis);
    style_textproperty_level.push_back(cc);
  } else {
    style_textproperty_level.push_back(0);    
  }
  style_textproperty_poetic.push_back((s.find (" poetic ") != string::npos));
  style_textproperty_note.push_back((s.find (" note ") != string::npos));
  style_textproperty_crossreference.push_back((s.find (" crossreference ") != string::npos));
  style_textproperty_verse.push_back((s.find (" verse ") != string::npos));

  s = usfm_get_style (style, "\\StyleType");
  if (s == "character") style_styletype.push_back(stCharacter);
  else if (s == "note") style_styletype.push_back(stNote);
  else if (s == "paragraph") style_styletype.push_back(stParagraph);
  // If no styletype was given, make it paragraph, the most common type
  // so that at least the content does not disappear into the nothing, 
  // but gets displayed.
  else {
    style_styletype.push_back(stParagraph);
    cerr << "Marker " << style_marker[style_marker.size() - 1] << " has no StyleType, so the default Paragraph has been assigned" << endl;
  }
 
  style_bold.push_back(usfm_style_present(style, "\\Bold"));

  style_italic.push_back(usfm_style_present(style, "\\Italic"));

  style_superscript.push_back(usfm_style_present(style, "\\Superscript"));

  s = usfm_get_style (style, "\\Endmarker");
  style_endmarker.push_back(s);

  s = usfm_get_style (style, "\\Justification");
  if (s == "center") style_justification.push_back(jCenter);
  else if (s == "left") style_justification.push_back(jLeft);
  else if (s == "right") style_justification.push_back(jRight);
  // If no justification is given, make it left, as default.
  else style_justification.push_back(jLeft);
 
  s = usfm_get_style (style, "\\FontSize");
  if (s.length()) {
    unsigned int fs = convert_to_int(s);
    style_fontsize.push_back(fs);
  } else {
    style_fontsize.push_back(12);
    cerr << "Marker " << style_marker[style_marker.size() - 1] << " has no FontSize, so the default of 12 has been assigned" << endl;
  }

  s = usfm_get_style (style, "\\LeftMargin");
  if (s.length()) {
    istringstream r(s);
    double lm;
    r >> lm;
    style_leftmargin.push_back(lm);
  } else {
    style_leftmargin.push_back(0);    
  }

  s = usfm_get_style (style, "\\RightMargin");
  if (s.length()) {
    istringstream r(s);
    double rm;
    r >> rm;
    style_rightmargin.push_back(rm);
  } else {
    style_rightmargin.push_back(0);    
  }

  s = usfm_get_style (style, "\\SpaceAfter");
  if (s.length()) {
    unsigned int sa = convert_to_int(s);
    style_space_after.push_back(sa);
  } else {
    style_space_after.push_back(0);
  }

  s = usfm_get_style (style, "\\SpaceBefore");
  if (s.length()) {
    unsigned int sb = convert_to_int (s);
    style_space_before.push_back(sb);
  } else {
    style_space_before.push_back(0);
  }

}


string Stylesheet::usfm_get_style (vector<string>& styles, const string& style) {
  // This function browses through the styles, and returns the value 
  // of style. E.g. when called with \Marker, it returns e.g. c
  // when \Marker c happens to be in the styles.
  string returnvalue;
  string s;
  string lowercasestyle = lowerCase(style);
  for (unsigned int i = 0; i < styles.size(); i++) {
    s = lowerCase (styles[i]);
    if (s.find(lowercasestyle) == 0) {
      s.erase(0, lowercasestyle.length());
      returnvalue = trim (s);
      break;
    }
  }
  return returnvalue;
}


bool Stylesheet::usfm_style_present (vector<string>& styles, const string& style) {
  // This function browses through the styles, and returns whether
  // style is present. E.g. when called with \Bold, and this style
  // is there, it returns true, else false.
  bool returnvalue = false;
  string s;
  string lowercasestyle = lowerCase(style);
  for (unsigned int i = 0; i < styles.size(); i++) {
    s = lowerCase (styles[i]);
    if (s.find(lowercasestyle) == 0) {
      returnvalue = true;
      break;
    }
  }
  return returnvalue;
}


bool Stylesheet::usfm_is_chapter (const string& usfm) {
  if (ChapterUsfm.empty()) {
    for (unsigned int i = 0; i < style_texttype.size(); i++) {
      if (style_texttype[i] == ttChapterNumber) {
	ChapterUsfm = style_marker[i];
	break;
      }
    }
    if (ChapterUsfm.empty()) {
      ChapterUsfm = "c";
      cerr << "No usfm for the chapter found, so the default value of c is being used" << endl;
    }
  }
  return (usfm == ChapterUsfm);
}


bool Stylesheet::usfm_is_id (const string& usfm) {
  if (IDUsfm.empty()) {
    for (unsigned int i = 0; i < style_texttype.size(); i++) {
      if (style_textproperty_book[i]) {
	IDUsfm = style_marker[i];
	break;
      }
    }
    if (IDUsfm.empty()) {
      IDUsfm = "id";
      cerr << "No usfm for the id of the book found, so the default value of id is being used" << endl;
    }
  }
  return (usfm == IDUsfm);
}


bool Stylesheet::usfm_is_verse (const string& usfm) {
  if (VerseUsfm.empty()) {
    for (unsigned int i = 0; i < style_texttype.size(); i++) {
      if (style_texttype[i] == ttVerseNumber) {
	VerseUsfm = style_marker[i];
	break;
      }
    }
    if (VerseUsfm.empty()) {
      VerseUsfm = "v";
      cerr << "No usfm for the verse found, so the default value of v is being used" << endl;
    }
  }
  return (usfm == VerseUsfm);
}


string Stylesheet::get_verse_usfm() {
  if (VerseUsfm.empty()) usfm_is_verse("");
  return VerseUsfm;
}


bool Stylesheet::UsfmIsRunningHeader (const string& usfm) {
  if (RunningHeaderUsfm.empty()) {
    for (unsigned int i = 0; i < style_texttype.size(); i++) {
      if (style_textproperty_publishable[i] && (style_texttype[i] == ttOther) && (style_textproperty_nonvernacular[i])) {
	RunningHeaderUsfm = style_marker[i];
	break;
      }
    }
    if (RunningHeaderUsfm.empty()) {
      RunningHeaderUsfm = "h";
      cerr << "No usfm for the running header of the book found, so using default value h" << endl;
    }
  }
  return (usfm == RunningHeaderUsfm);
}


int Stylesheet::UsfmIndex (const string& usfm) {
  for (unsigned int i = 0; i < style_texttype.size(); i++) {
    if (usfm == style_marker[i])
      return i;
  }
  return -1;			// USFM does not exist: return -1
}


unsigned int Stylesheet::Fontsize (const unsigned int index) {
  // This function does not return the size of the font in points, as
  // given in the stylesheet, but instead it converts it to
  // size numbers as used in html.
  unsigned int i = style_fontsize [index];
  unsigned int returnvalue = 3;
  if (i > 10) returnvalue = 3;
  if (i > 12) returnvalue = 4;
  if (i > 14) returnvalue = 5;
  if (i > 16) returnvalue = 6;
  if (i > 20) returnvalue = 7;
  return returnvalue;
}


unsigned int Stylesheet::VersetextIndex () {
  unsigned int returnvalue = 0;
  for (unsigned int i = 0; i < style_texttype.size(); i++) {
    if (style_texttype[i] == ttVerseText) {
      returnvalue = i;
      break;
    }
  }
  return returnvalue;
}


unsigned int Stylesheet::VersenumberIndex () {
  unsigned int returnvalue = 0;
  for (unsigned int i = 0; i < style_texttype.size(); i++) {
    if (style_texttype[i] == ttVerseNumber) {
      returnvalue = i;
      break;
    }
  }
  return returnvalue;
}


bool Stylesheet::UsfmBold (const unsigned int index) {
  return style_bold [index];
}


bool Stylesheet::UsfmCenter (const unsigned int index) {
  return (style_justification[index] == jCenter);
}


bool Stylesheet::UsfmItalics (const unsigned int index) {
  return style_italic [index];
}


bool Stylesheet::UsfmNonvernacular (const unsigned int index) {
  return style_textproperty_nonvernacular [index];
}


bool Stylesheet::GetUsfmNewParagraph (const unsigned int index) {
  return (style_styletype [index] == stParagraph);
}


string Stylesheet::GetEndmarker (const unsigned int index) {
  return style_endmarker [index];
}


