/*
** Copyright (C) 2003 The Free Software Foundation.
**  
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 2 of the License, or
** (at your option) any later version.
**  
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**  
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software
** Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
**  
*/

#ifndef INCLUDED_STYLESHEET_H
#define INCLUDED_STYLESHEET_H

#include "libraries.h"

class Stylesheet {
 public:
  Stylesheet ();
  ~Stylesheet ();
  void readsheet (const string& filename);
  bool usfm_is_chapter (const string& usfm);
  bool usfm_is_id (const string& usfm);
  bool usfm_is_verse (const string& usfm);
  string get_verse_usfm();
  bool UsfmIsRunningHeader (const string& usfm);
  int UsfmIndex (const string& usfm);
  unsigned int Fontsize (const unsigned int index);
  unsigned int VersetextIndex ();
  unsigned int VersenumberIndex ();
  bool UsfmBold (const unsigned int index);
  bool UsfmCenter (const unsigned int index);
  bool UsfmItalics (const unsigned int index);
  bool UsfmNonvernacular (const unsigned int index);
  bool GetUsfmNewParagraph (const unsigned int index);
  string GetEndmarker (const unsigned int index);
 private:
  void LoadStylesheet (vector<string>& style);
  string usfm_get_style (vector<string>& styles, const string& style);
  bool usfm_style_present (vector<string>& styles, const string& style);
  string ChapterUsfm;
  string IDUsfm;
  string VerseUsfm;
  string RunningHeaderUsfm;

  /* These are the texttypes currently supported in Paratext 6
  \TextType ChapterNumber
  \TextType NoteText
  \TextType Other
  \TextType Section
  \TextType Title
  \TextType VerseNumber
  \TextType VerseText
  */
  typedef enum {ttChapterNumber, ttNoteText, ttOther, ttSection, ttTitle, ttVerseNumber, ttVerseText} texttypes;

  vector<string> style_marker;        // E.g. \Marker c
  vector<texttypes> style_texttype;   // E.g. \TextType ChapterNumber

  /* These are the textproperties current found in Paratext 6.

  chapter
  paragraph
  nonpublishable or publilshable (or nothing at all)
  nonvernacular or vernacular (or nothing at all)
  book
  level_1 or level_2 or level_3 or level_4 (or no level indicator at all)
  poetic
  note
  crossreference
  verse

  More than one property can occur at once, as can be seen
  in the list below:

  \TextProperties chapter
  \TextProperties paragraph nonpublishable nonvernacular
  \TextProperties paragraph nonpublishable nonvernacular book
  \TextProperties paragraph publishable nonvernacular
  \TextProperties paragraph publishable vernacular
  \TextProperties paragraph publishable vernacular level_1
  \TextProperties paragraph publishable vernacular level_2
  \TextProperties paragraph publishable vernacular level_4
  \TextProperties paragraph publishable vernacular poetic
  \TextProperties paragraph publishable vernacular poetic level_1
  \TextProperties paragraph publishable vernacular poetic level_2
  \TextProperties paragraph publishable vernacular poetic level_3
  \TextProperties publishable vernacular
  \TextProperties publishable vernacular note
  \TextProperties publishable vernacular note crossreference
  \TextProperties verse
  */

  vector<bool> style_textproperty_chapter;
  vector<bool> style_textproperty_paragraph;
  vector<bool> style_textproperty_publishable;
  vector<bool> style_textproperty_nonpublishable;
  vector<bool> style_textproperty_nonvernacular;
  vector<bool> style_textproperty_vernacular;
  vector<bool> style_textproperty_book;
  vector<unsigned int> style_textproperty_level;
  vector<bool> style_textproperty_poetic;
  vector<bool> style_textproperty_note;
  vector<bool> style_textproperty_crossreference;
  vector<bool> style_textproperty_verse;


  /* These are the styletypes currently found in Paratext 6.
     \StyleType Character
     \Styletype Note
     \StyleType Paragraph
  */

  typedef enum {stCharacter, stNote, stParagraph} styletypes;
  vector<styletypes> style_styletype;   // E.g. \StyleType Paragraph

  // Dealing with \Bold
  vector<bool> style_bold;

  // Dealing with \Italic
  vector<bool> style_italic;

  // Dealing with \Superscript
  vector<bool> style_superscript;

  // Dealing with the endmarker, e.g. \Endmarker bw*
  vector<string> style_endmarker;

  /* Dealing with justification. This is at present supported in P6:
     \Justification Center
     \Justification Left
     \Justification Right
  */
  typedef enum {jCenter, jLeft, jRight} justification;
  vector<justification> style_justification;

  /* Deal here with FirstLineIndent
     E.g. \FirstLineIndent -.125   # 1/4 inch second line indent
  */
  vector<double> style_first_line_indent;

  // Deal with the FontSize.
  vector<unsigned int> style_fontsize; // E.g. \FontSize 9

  // Deal with the LeftMargin
  vector<double> style_leftmargin; // E.g. \LeftMargin .25

  // Deal with the RightMargin
  vector<double> style_rightmargin; // E.g. \RightMargin .25

  // Deal with the SpaceAfter
  vector<unsigned int> style_space_after; // E.g. \SpaceAfter 2

  // Deal with the SpaceBefore
  vector<unsigned int> style_space_before; // E.g. \SpaceBefore 16

  // IMPORTANT: When adding new variables, have them cleared in function readsheet.

};

#endif


