/*
** Copyright (C) 2003 The Free Software Foundation.
**  
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 2 of the License, or
** (at your option) any later version.
**  
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**  
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software
** Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
**  
*/


#include "libraries.h"
#include "utilities.h"
#include <libgen.h>
#include <gtkmm.h>
 


string trim(const string& s) {
  if (s.length() == 0)
    return s;
  unsigned int beg = s.find_first_not_of(" \t");	// Strip spaces, tabs and new lines
  unsigned int end = s.find_last_not_of(" \t\n\r");
  if (beg == string::npos)	// No non-spaces
    return "";
  return string (s, beg, end - beg + 1);
}


string number_in_string (const string& str) {
  // Looks for and returns a positive number in a string.
  string output = str;
  #define MY_NUMBERS "0123456789"
  output.erase(0, output.find_first_of(MY_NUMBERS));
  unsigned int end_position = output.find_first_not_of(MY_NUMBERS);
  #undef MY_NUMBERS
  if (end_position != string::npos) {
    output.erase(end_position, output.length());
  }
  return output;
}


string upperCase(const string& s) {
// Make an uppercase copy of s
  string upper(s);
  for(size_t i = 0; i < s.length(); ++i)
    upper[i] = toupper(upper[i]);
  return upper;
}


string lowerCase(const string& s) {
// Make a lowercase copy of s
  string lower(s);
  for(size_t i = 0; i < s.length(); ++i)
    lower[i] = tolower(lower[i]);
  return lower;
}


string remove_spaces(const string& s) {
  string s2 = s;
  unsigned int spacepos = s2.find(" ");
  while (spacepos != string::npos) {
    s2.erase(spacepos,1);
    spacepos = s2.find(" ");
  }
  return s2;
}


int create_directory (const string& dir) {
  // Creates directory, with the parents, if need be.
  // Function mkdir could be used (see man 2 mkdir), but this does not allow for 
  // the creation of the parent directories. The core utility mkdir provides
  // this functionality, so is preferred, and used here.
  string s;
  s = "mkdir -p '" + dir + "'";
  return system (s.c_str());
}


unsigned int convert_to_int (const string& str) {
  unsigned int i = 0;
  istringstream r(str);
  r >> i;
  return i;
}


string convert_to_string (int i) {
  ostringstream r;
  r << i;
  return r.str();
}


string convert_to_string (unsigned int& i) {
  int i2 = i;
  return convert_to_string(i2);
}


string convert_to_string (bool b) {
  if (b) return "1";
  else return "0";
}


bool convert_to_bool (const string& s) {
  if (s == "") return false;
  if (s == "0") return false;
  if (s == "F") return false;
  else return true;
}


void write_lines (const string& file, vector<string>& lines) {
  WriteText wt (file);
  for (unsigned int i = 0; i < lines.size(); i++) {
    wt.out << lines[i] << endl;
  }
}


string temporary_file(const string& filename) {
  return ((Glib::build_filename(Glib::get_tmp_dir(), filename)));
}


bool program_is_running (const string& commandline) {
  /* 
     Returns true if the program given on "commandline"
     is running, else false.
  */

  // Run the process status program.
  FILE * fd;
  fd = popen ("ps -ax", "r");
  if (!fd) {
    return false;
  }
  // Read the results of it.
  int i = 0;
  char buf[1024];
  while (fgets (buf, sizeof(buf), fd) != NULL) {
    if (strstr (buf, commandline.c_str()) != NULL)
      i++;
  }
  int result;
  result = pclose(fd);
  // Return true if one or more of the commandlines were found.
  // That is, return true if the program runs.
  return (i > 0);
}


bool string_in_file (const string& filename, const string& line) {
  /* 
     Returns true if the string "line" is found in file "filename".
   */
  string command = "grep --quiet '" + line + "' '" + filename + "' > /dev/null 2>&1";
  int i = system(command.c_str()); // Will return exitstatus 0 if a match was found.
  return (i == 0);
}


ReadDirectory::ReadDirectory (const string& path, const string& prefix, const string& suffix) {
  // Reads the files in directory "path" that end on "suffix".
  try {
    Glib::Dir dir (path);
    vector<string> entries (dir.begin(), dir.end());
    for (unsigned int i = 0; i < entries.size(); i++) {
      if (Glib::str_has_suffix(entries[i], suffix))
	if (Glib::str_has_prefix(entries[i], prefix))
	  files.push_back (entries[i]);
    }
  }
  catch(...) {
  }
}


ReadDirectory::~ReadDirectory() {
}


ReadText::ReadText (const string& file, bool silent) {
  // Reads the text and stores it line by line, trimmed, into "Lines".
  // If "silent" is true, then no exception will be thrown in case of an error.
  ifstream in(file.c_str());
  if(!in) {
    if (!silent) {
      cerr << "Error opening file " << file << endl;
      throw;
    }
    return;
  }
  string s;
  while (getline(in, s)) 
    lines.push_back(trim(s));
}


ReadText::~ReadText () {
}


WriteText::WriteText (const string& file, bool append) {
  /*
    This opens a textfile for writing.
    If 'append' is set, it appends the output to an existing file.
   */
  if (append) {
    out.open (file.c_str(), ios_base::out | ios_base::app);
  } else {
    out.open (file.c_str());
  }
  if(!out) {
    cerr << "Error creating file " << file << endl;
    throw 1;
  }
}


WriteText:: ~WriteText() {
}


